#! /usr/bin/env python

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


__doc__ = """

This module is ressponsible for parsing xml nufw firewall desccription, and mapping it into objects.

""" 
import sys
import xml.dom.minidom
import IPy
import copy

ZONE_EXT="external"
ZONE_INT="internal"
ZONE_VPN="vpn"

class UnknownVpn(Exception):
    def __init__(self,mark):
	self.mark=mark
	
    def __str__(self):
	return 'unknown vpn with mark %s'%(self.mark)
    
class Net:
    def __init__(self,name,addr,zone='',type='',mark='',remote=''):
	if(zone==''):
	    self.__zone=ZONE_INT
	else:
	    self.__zone=zone
	self.__conns=[]
	if addr == '': addr = '0.0.0.0/0'
	self.__addr=IPy.IP(addr)
	self.__name=name
	self.__type=type
	self.__mark=mark
	self.__remote = remote
	
    def add_conn(self,dir,fwid,iface,*args):
	h={'direct':dir, 'fwid':fwid, 'iface':iface}
	if(len(args)==1):
	    h['router']=args[0]
	#print h
	self.__conns.append(h)
	
    def get_conn(self,fwid):
	for conn in self.__conns:
	    if conn['fwid']==fwid:
		return conn
	
    def iter_conns(self):
	for index in range(0,len(self.__conns)):
	    yield self.__conns[index]
    
    #getters
    def netmask(self):
	return self.__addr.netmask()

    def ip(self):
	return str(self.__addr)

    def addr(self):
	return self.__addr
    
    def name(self):
	return self.__name
   
    def zone(self):
	return self.__zone
	
    def type(self):
	return self.__type
    
    def mark(self):
	return self.__mark
   
    def remote(self):
	return self.__remote
    
    def __str__(self):
	s='name: %s, addr:%s, zone:%s' %(self.__name,str(self.__addr), self.__zone)
	return s
    def __repr__(self):
	return str(self)
#End class Net

class Interface:
    def __init__(self,id,name,addr):
	self.__id=id
	self.__net=None
	self.__addr=IPy.IP(addr)
	self.__name=name
	self.__distnets=[]
	self.__nets=[]
	
    def name(self) :
	return self.__name

    def set_net(self,net):
	self.__net=net
	
    def add_distnet(self,net):
	self.__distnets.append(net)
    
    def add_net(self,net):
	self.__nets.append(net)
    
    def distnets(self):
	for n in self.__distnets:
	    yield n
    
    def allnets(self):
	for n in self.__nets:
	    yield n
    
    def netmask(self):
	if(type(self.__net)==type(None)):
	    return ''
	return self.__net.netmask()

    def addr(self):
	return repr(self.__addr)
    
    def id(self):
	return self.__id
    
    def net(self):
	return self.__net
    def deb_conf(self):
	s="""
auto %s
iface %s inet static
    address %s
    netmask %s""" %(self.__name, self.__name, self.__addr,self.netmask())
	return s

    
#End class Interface
class RuleList(object):
    def __init__(self):
	self.rules = []
	self.bad = 0

    def is_bad(self):
	return self.bad == 1
    
    def add_rule(self,rule):
	self.rules.append(rule)
    
class FW:
    def __init__(self,id):
	self.__ifaces=[]
	self.__nets=[]
	self.__id=id
	self.__hrules={}
	self.__allrules=[]
	self.__chains=[]
	self.__ext_net=None
	self.__vpns={}
    
    def add_iface(self,iface):
	self.__ifaces.append(iface)
	
    def ifaces(self):
	for i in self.__ifaces:
	    yield i

    def add_net(self,net):
	self.__nets.append(net)
	#self.__chains=self.chains(all=1)
	if net.zone()==ZONE_EXT:
	    self.__ext_net=net
	if net.type()=='mark':
	    self.__vpns[net.mark()]=net
	
	    

    
    def deb_conf(self):
	s="""
# /etc/network/interfaces -- configuration file for ifup(8), ifdown(8)
# 
# This file has been generated by nufw 

auto lo 
iface lo inet loopback
	"""
	for myiface in self.__ifaces:
	    s+="""%s
	    """ %(myiface.deb_conf())
	return s

    def routes(self):
	s=''
	for iface in self.__ifaces:
	    for n in iface.distnets():
		h=n.get_conn(self.__id)
		#print 'route ',h
		s+="%s\t%s\n" %(str(n.addr().net()),h['router'])
	return s
	
    def chain(self,ip_src,ip_dst,zone_src='',zone_dst='',mark_src='',mark_dst=''):
	chp1=''
	chp2=''
	find1=0
	find2=0
	#TODO: get net() instead of net().name(), return hash
	#print 'in chain |%s| |%s| |%s| |%s|' %(ip_src,zone_src,zone_dst,ip_dst)
	if mark_src!='':
	    #print 'chain mar_src', mark_src
	    if not self.__vpns.has_key(mark_src):
		raise UnknownVpn(mark_src)
	    chp1=self.__vpns[mark_src]
	    find1=1
	if mark_dst!='':
	    #print 'chain mar_src', mark_dst
	    if not self.__vpns.has_key(mark_dst):
		raise UnknownVpn(mark_dst)
	    chp2=self.__vpns[mark_dst]
	    find2=1

	if ip_dst=='' and zone_dst=='external':
	    chp2=self.__ext_net
	    find2=1
	    ip_dst='0.0.0.0'
	if ip_src=='' and zone_src=='external':
	    chp1=self.__ext_net
	    find1=1
	    ip_src='0.0.0.0'
	for iface in self.__ifaces:
	    if(iface.net().zone()==ZONE_INT and ip_src in iface.net().addr()):
		chp1=iface.net()
		find1=1
	    if(iface.net().zone()==ZONE_INT and ip_dst in iface.net().addr()):
		chp2=iface.net()
		find2=1
	    for n in iface.distnets():
		if(n.zone()==ZONE_INT and ip_src in n.addr()):
		    chp1=n
		    find1=1
		if(n.zone()==ZONE_INT and ip_dst in n.addr()):
		    chp2=n
		    find2=1
	if(find1==0):
	    chp1=self.__ext_net
	if(find2==0):
	    chp2=self.__ext_net
	return {'chain':chp1.name()+'-'+chp2.name(), 'netsrc':chp1, 'netdst':chp2}
	
    def chains(self,all=0):
	l=[]
	for iface in self.__ifaces:
	    #print (all==0 and iface.net().zone()!=ZONE_EXT) or all==1
	    if((all==0 and iface.net().zone()==ZONE_INT) or all==1):
		l.append({'name':iface.net().name(), 'net':iface.net(),'zone':iface.net().zone(),'iface':iface})
	    for n in iface.distnets():
		if((all==0 and iface.net().zone()==ZONE_INT) or all==1):
		    l.append({'name':n.name(), 'net':n,'zone':iface.net().zone(),'iface':iface})
	#print "in chains ",l
	def _fun(l1,l2):
	    res=[]
	    for u1 in l1:
		for u2 in l2:
		    if(u1['name']!=u2['name']):
			res.append({'name':u1['name']+'-'+u2['name'], 'net_src':u1['net'], 'net_dst':u2['net'],'if_src':u1['iface'],'if_dst':u2['iface']})
			if not self.__hrules.has_key(u1['name']+'-'+u2['name']):
			    self.__hrules[u1['name']+'-'+u2['name']] = RuleList()
	    return res
	return _fun(l,l)

    def entreprise_nets(self):
	l=[]
	for iface in self.__ifaces:
	    n=iface.net()
	    if(n.zone()==ZONE_INT):
		l.append(str(n.addr()))
	    for n in iface.distnets():
		if(n.zone()==ZONE_INT):
		    l.append(str(n.addr()))
	return l

    def add_rule(self, r):
	"""Add a rule to the firewall into the right chain
	"""
	#print 'in add_rule',r.aclid,r.elt_src(), r.elt_dest()
	ch=self.chain(r.elt_src().get('numnet'),r.elt_dest().get('numnet'),zone_src=r.elt_src().get('zone'),zone_dst=r.elt_dest().get('zone'),mark_src=r.elt_src().get('mark'),mark_dst=r.elt_dest().get('mark'))
	#print 'in add_rulesrc',r.aclid,ch['netsrc'],ch['netsrc'].addr()
	#r.set_netsrc(copy.deepcopy(ch['netsrc']))
	#r.set_netdst(copy.deepcopy(ch['netdst']))
	r.set_netsrc(ch['netsrc'])
	r.set_netdst(ch['netdst'])
	r.snetsrc=str(ch['netsrc'].addr())
	r.snetdst=str(ch['netdst'].addr())
	#print 'in add_rule ', ch, r.elt_src(), r.elt_dest()
	if(not self.__hrules.has_key(ch['chain'])):
	    #self.__hrules[ch['chain']]=[]
	    self.__hrules[ch['chain']] = RuleList()
	    #print type(self.__hrules[ch['chain']])
	    #if r.snetsrc == r.snetdst: 
	    self.__hrules[ch['chain']].bad = 1
		#print self.__hrules[ch['chain']]
	#print self.__hrules[ch['chain']]['rules']
	self.__hrules[ch['chain']].add_rule(r)
	#(self.__hrules[ch['chain']]['rules']).append(r)
	#print len(self.__hrules[ch['chain']]['rules'])
	self.__allrules.append(r)

    def hrules(self):
	return self.__hrules

    def allrules(self):
	return self.__allrules
#End class FW


def parse(file):
    fwhash={}
    doc=xml.dom.minidom.parse(file)
    fwl=[]
    netl={}
    eltsfw=doc.getElementsByTagName("fw")
    eltnets=doc.getElementsByTagName('nets')
    eltsnet=eltnets[0].childNodes
    for eltnet in eltsnet:
	if(eltnet.nodeType==xml.dom.minidom.Node.ELEMENT_NODE and eltnet.nodeName=="net"):
	    #print eltnet
	    #print "\t",eltnet.nodeName
	    #print "\tname ", eltnet.getAttribute('name')
	    #print "\taddr ", eltnet.getAttribute('addr')
	    #print "net mark",eltnet.getAttribute('mark')
	    cnet=Net(eltnet.getAttribute('name'),eltnet.getAttribute('addr'),zone=eltnet.getAttribute('zone'),type=eltnet.getAttribute('type'),mark=eltnet.getAttribute('mark'),remote=eltnet.getAttribute('remote'))
	    if(eltnet.hasChildNodes()):
		#print "\thas childs"
		for conn in eltnet.getElementsByTagName('connection'):
		    fwid=conn.getAttribute('fwid')
		    direct=conn.getAttribute('direct')
		    iface=conn.getAttribute('iface')
		    sd=''
		    if(direct=='1'): 
			sd='directe'
			cnet.add_conn(direct,fwid,iface)
		    else: 
			sd='indirecte'
			cnet.add_conn(direct,fwid,iface,conn.getAttribute("router"))
		#put new found net into global nets list
		    if(not netl.has_key(fwid)): netl[fwid]=[cnet]
		    else: netl[fwid].append(cnet)
    for elt in eltsfw:
	#print elt
	fwid=elt.getAttribute('id')
	#print "id ",fwid
	#print "type ",elt.getAttribute('type')
	#print "name ",elt.getAttribute('name')
	eltsinterface=elt.getElementsByTagName('interface')
	myfw=FW(fwid)
	for interface in eltsinterface:
		#print "\t", interface
		#print "\tname", interface.getAttribute("name")
		myiface=Interface(interface.getAttribute('id'),interface.getAttribute("name"),interface.getAttribute('addr'))
		#print "\t",myiface
		#print "\t\tconn %s pour %s" %(sd,fwid)
		myfw.add_iface(myiface)
		#print netl[fwid]
		for net in netl[fwid]:
		    #print "\tnetnet",net
		    #print "\t",net.ip()
		    for conn in net.iter_conns():
		    	#print "\tconn ",conn
			if (conn['fwid']==fwid and conn['iface']==myiface.id()):
			    if conn['direct']=='1':
				myiface.set_net(net)
			    else: 
				#print "adding ",conn
				myiface.add_distnet(net)
			    myiface.add_net(net)
		    myfw.add_net(net)
		#print myiface.deb_conf()
	fwhash[fwid]=myfw
    return fwhash

    
