#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


import fw , nuacl , os
import string
from IPy import IP
from cStringIO import StringIO


IPTABLES="$ipt"

D_ACCEPT='ACCEPT'
D_REJECT='REJECT'
D_DROP='DROP'
D_ULOG='ULOG'
D_LOG='LOG'
D_DEFAULT=D_DROP


def gen_rule(*args):
    #s=IPTABLES+' '+string.join(args)
    #return s
    return IPTABLES+' '+string.join(args)
    
def gen_rule2(l):
    return gen_rule(*l)

def make_target(s):
    if s=='accept':
	return D_ACCEPT
    if s=='reject':
	return D_REJECT
    if s=='ulog':
	return D_ULOG
    if s=='log':
	return D_LOG
    return D_DROP

def _make_syn(target,h):
    """returns chain that feets well to the target of the iptables command
	example: for accept target: we want to match syn packets, the others are matched by the
	RELATED/ESTABLISHED rules
    """
    s=target
    res=''
    if s == D_ACCEPT:
        res +='-m state --state NEW'
        if h.has_key('proto') and h['proto']=='tcp':
	    res+=' --syn'
    return res

class FWipt:
    """Used to generate iptables rules from a firewall object that contains rules
    """
    def __init__(self,thefw,logtype='log'):
	self.__fw=thefw
	self.__logtype='LOG'
	self.__logprefix='--log-prefix'
	self.__logspace=' ' #one space
	if logtype=='ulog': 
	    self.__logtype='ULOG'
	    self.__logprefix='--ulog-prefix'
	    self.__logspace=''
   
    def dumb_rules(self,rules,chain):
	res = ['#ERROR_DISPATCH %s%s'%(chain,os.linesep)]
	sacl = ''
	for r in rules:
	    stmp=r.aclname
	    if(stmp!=sacl):
		sacl=stmp
		sc='#'+sacl+' (acl '+r.aclid+')'+os.linesep
		res.append(sc)
	    s=str(RulePrintIptSimple(r,chain=chain,add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision)))+os.linesep
	    res.append('#'+s)
	return res
    dumb_rules = classmethod(dumb_rules)
    
    def create_ipt_chains(self):
	"""Generate rules to create firewall chains: DMZ-LAN
	"""
	l=self.__fw.chains(all=1)
	s=''
	s2=''
	acc=[]
	for n in l:
	    st=IPTABLES+' -N '+n['name']+os.linesep
	    if not st in acc:
		s+=st
	    acc.append(st)
	return s

    def connect_chains(self,rescue=0):
	"""Generate rules that connect firewall streams into chains
	"""
	l=self.__fw.chains(all=1)
	s=''
	svpn=''
	estrel=''
	sinternet=''
	if rescue==1: #backup mode: we need only one global cmd for established/related packets
	    estrel=gen_rule('-A FORWARD -m state --state ESTABLISHED,RELATED -j ACCEPT')+os.linesep
	else:
	    estrel=gen_rule('-A FORWARD -m state --state RELATED -j ACCEPT')+os.linesep

	for n in l:
	    #use gen_rule2 and insert the right elements in list switch zone
	    #
	    has_vpn = False
	    has_internet = False
	    t=['-A FORWARD']
	    nsrc=n['net_src']
	    ndst=n['net_dst']
	    if nsrc.zone()==fw.ZONE_INT or nsrc.zone()==fw.ZONE_VPN:
		t.extend(['-s',str(nsrc.ip())])
	    if ndst.zone()==fw.ZONE_INT or ndst.zone()==fw.ZONE_VPN:
		t.extend(['-d',str(ndst.ip())])
	    if nsrc.zone()==fw.ZONE_VPN or ndst.zone()==fw.ZONE_VPN:
		themark=nsrc.mark()
		if themark=='': themark=ndst.mark()
		t.extend(['-m mark --mark',themark])
		has_vpn=True
	    if nsrc.zone()==fw.ZONE_EXT or ndst.zone()==fw.ZONE_EXT:
		has_internet = True
	    t.extend(['-i',n['if_src'].name(),'-o',n['if_dst'].name(),'-j',n['name']])
	    #st=gen_rule('-s',str(n['net_src'].ip()),'-d',str(n['net_dst'].ip()),'-i',n['if_src'].name(),'-o',n['if_dst'].name(),'-j',n['name'])
	    st=gen_rule2(t)+os.linesep
	    if nsrc.zone()==fw.ZONE_VPN and ndst.zone()==fw.ZONE_VPN:
		st='' 
	    if has_vpn is True: #we need vpn related flux to be first
		svpn+=st
	    else: 
		if has_internet is True and has_vpn is False:
		    sinternet+=st
		else:
		    s+=st
	invalid=gen_rule('-A FORWARD -m state --state INVALID -j DROP')+os.linesep*2
	log_forward=gen_rule('-A FORWARD -j %s %s "DEFAULT_FWD_DROP%s" '%(self.__logtype,self.__logprefix,self.__logspace))+os.linesep
	drop_forward=gen_rule('-A FORWARD -j DROP')+os.linesep
	
	return estrel+invalid+svpn+s+sinternet+log_forward+drop_forward+os.linesep

    def gen_rules_chain(self,chain,rescue=0):
	s=''
	s2=''
	s3=''
	file_s = StringIO()
	file_s2 = StringIO()
	l=[os.linesep*2]
	l2=[]
	l3=[]
	l_eld = [] #list for established log drop rules: put it at the end to avoid inverse chain bug
	rl = self.__fw.hrules()[chain]
	#print chain
	#print type(self.__fw.hrules()[chain])
	#print self.__fw.hrules()[chain]
	#if rl.is_bad():
	    #print chain
	    #print rl.bad
	lch=self.__fw.hrules()[chain].rules
	if rl.is_bad():
	    return FWipt.dumb_rules(lch,chain)
	lch.sort()
	estrel=''
	sacl=''
	stmp=''
	comment=0
	for r in lch:
	    stmp=r.aclname
	    comment_rule=''
	    if(stmp!=sacl):
		sacl=stmp
		sc=os.linesep+'#'+sacl+' (acl '+r.aclid+')'+os.linesep
		comment=1
	    else: comment=0
	    if r.from_nufw():# and r.proto().get('proto') in ['tcp']:#generate QUEUE cmds as in the Misc Article
		if r.proto().get('proto') != 'tcp':#we comment bogus rules to avoid iptables errors and let client learn
		    comment_rule='#'
		    l.append('#ERROR_PROTO_AUTH'+os.linesep)
		if rescue==0:
		    if comment==1: 
			#s+=sc
			#file_s.write(sc)
			l.append(sc)
		    inv=r.invert()
		    invchain=self.__fw.chain(r.elt_dest().get('numnet'),r.elt_src().get('numnet'),r.elt_dest().get('zone'),r.elt_src().get('zone'),mark_src=r.elt_src().get('mark'),mark_dst=r.elt_dest().get('mark'))
		    sinv=str(RulePrintIptSimple(inv,chain=invchain['chain'],target='QUEUE',insert='-A',add='-m state --state ESTABLISHED --tcp-flags SYN,ACK,RST,FIN SYN,ACK'))+os.linesep
		
		    #s+=sinv
		    #file_s.write(sinv)
		    rst=RulePrintIptSimple(r,chain=chain,target='QUEUE',add='-m state --state ESTABLISHED --tcp-flags RST RST')
		    fin=RulePrintIptSimple(r,chain=chain,target='QUEUE',add='-m state --state ESTABLISHED --tcp-flags FIN FIN')
		    #s+=str(rst)+os.linesep
		    #s+=str(fin)+os.linesep
		    #file_s.write(str(rst)+os.linesep)
		    #file_s.write(str(fin)+os.linesep)
		    l.extend([comment_rule+sinv,comment_rule+str(rst)+os.linesep,comment_rule+str(fin)+os.linesep])
		    ssyn=str(RulePrintIptSimple(r,chain=chain,target='QUEUE',add='-m state --state NEW --syn'))+os.linesep
		    if comment==1: 
			#s2+=sc
			#file_s2.write(sc)
			l2.append(sc)
		    #s2+=ssyn
		    #file_s2.write(ssyn)
		    l2.append(comment_rule+ssyn)
		else:
		    srescue=str(RulePrintIptSimple(r,chain=chain,add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision)))+os.linesep
		    if comment==1: 
			#s2+=sc
			#file_s2.write(sc)
			l2.append(sc)
		    #s2+=srescue
		    #file_s2.write(srescue)
		    l2.append(srescue)
	    else:#generate normal iptables cmds: target depends on _decision_ attribut
		std=str(RulePrintIptSimple(r,chain=chain,add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision)))
		if comment==1: 
		    #s2+=sc
		    #file_s2.write(sc)
		    l2.append(sc)
		#s2+=std+os.linesep
		#file_s2.write(std+os.linesep)
		l2.append(std+os.linesep)
	relaccept=''
	if(rescue==0):#backup mode: don't need REL/EST command per chain
	    relaccept=gen_rule('-A',chain,'-m state --state ESTABLISHED','-j ACCEPT')+os.linesep
	
	log=gen_rule('-A',chain,'-j',self.__logtype,self.__logprefix,'"%s%s"'%(chain,self.__logspace))+os.linesep
	drop=gen_rule('-A',chain,'-j DROP')

	l.extend(l2)
	l_eld.extend([relaccept,log,drop])
	#return (os.linesep*2)+s+s2+relaccept+log+drop
	#return (os.linesep*2)+file_s.getvalue()+file_s2.getvalue()+relaccept+log+drop
	return (l, l_eld) #return rules from acls and end chain rules separately to avoid inverse chain bug
     
    def gen_rules(self,rescue=0):
	s=''
	estrel=''
	l = []
	l_eld = []
	sep = os.linesep*2
	for ch in self.__fw.hrules().keys():
	    #s+=self.gen_rules_chain(ch,rescue=rescue)
	    #l.extend(self.gen_rules_chain(ch,rescue=rescue))
	    (a,b) = self.gen_rules_chain(ch,rescue=rescue)
	    l.extend(a)
	    l_eld.extend(b)
	    l.append(sep)
	    l_eld.append(sep)
	    #s+=os.linesep*2
	#return estrel+s
	l.extend(l_eld)
	return l
   
	
    def gen_vpn_rules(self):
	l = []
	for iface in self.__fw.ifaces():
	    for n in iface.allnets():
		if n.zone() == fw.ZONE_VPN:
		    l.append('#%s%s'%(n.name(),os.linesep))
		    l.append(gen_rule('-A PREROUTING','-t mangle','-i',iface.name(),'-p 50','-s',n.remote(),'-j MARK --set-mark',n.mark())+os.linesep)
		    l.append(gen_rule('-A PREROUTING','-t mangle','-d',str(n.addr()),'-j MARK --set-mark',n.mark())+os.linesep)
	return l
	
#End class FWipt

class RulePrint:
    def __init__(self,r,fw='',net_from='',net_to=''):
	self.rule=r
	self.fw=fw
	self.ef=self.rule.elt_src()
	self.et=self.rule.elt_dest()
	self.do_vpn = True
	#if net_from!='':
	    #self.net_from=net_from
	#else
	    #self.net_from=self.ef.get('numnet')
	   # 
	#if net_to!='':
	    #self.net_to=net_to
	#else
	    #self.net_to=self.et.get('numnet')

    def __str__(self):
	pass
    
class RulePrintIptSimple(RulePrint):
    """produces a iptables command form an acl rule
    """
    #TODO: take chain and target in argument
    def __init__(self, r,fw='',chain='',target='DROP',net_from='',net_to='',add='',insert=''):
	#super(RulePrintIpt, self).__init__(r)
	RulePrint.__init__(self, r, fw='',net_from=net_from,net_to=net_to)
	self.target=target
	self.chain=chain
	self.add=add
	self.insert='-A'
	if insert != '':
	    self.insert=insert
	#if netsrc is not None: self.netsrc=netsrc
	#else: 
	self.netsrc=self.rule.netsrc
	#if netdst is not None: self.netdst=netdst
	#else: 
	self.netdst=self.rule.netdst
	self.ulog_prefix=self.chain
	#self.dir=dir
	
    def _net_is_inclued_in(self,inner,net):
	"""check if inner is inclued in net
	"""
	#print 'inclued %s' %(self.rule.aclid)
	#print '_net_is_inclued_in:',inner,net,type(net),self.rule.snetsrc,self.rule.aclid
	if net is None: return False
	if inner.find('/')==-1: return False
	if net.zone()=='external': return False
	if net.zone()=='vpn': return False
	ntest=IP(inner)
	bool = ntest in net.addr()
	#print '%s %s in %s : %s'%(self.rule.aclid,inner,str(net.addr()),str(bool))
	if ntest == net.addr(): return False
	if ntest in net.addr():
	    #print '_net_is_inclued return true'
	    return True
	return False
    def set_ulog_prefix(self,s):
	self.ulog_prefix=s
    
    def __str__(self):
	#return "str RulePrintIpt"
	ef=self.rule.src().elts[0]
	et=self.rule.dest().elts[0]
	pr=self.rule.proto()
	t=[self.insert, self.chain]
	proto=''
	dport=''
	sport=''
	icmptype=''
	sip=''
	dip=''
	mark=''
	print_sip=False
	print_dip=False
	if pr.get('proto')!='':
	    proto=pr.get('proto')
	#if pr.get('proto')!='':
	    #proto=et.get('proto')
	if proto!='':
	    t.extend(['-p',proto])

	if pr.get('icmptype')!='':
	    icmptype=pr.get('icmptype')
	if pr.get('dport')!='':
	    dport=pr.get('dport')
	#if pr.get('dport')!='':
	    #dport=pr.get('dport')
	if pr.get('sport')!='':
	    sport=pr.get('sport')
	#if et.get('sport')!='':
	    #sport=et.get('sport')
	if et.get('mark')!='':
	    mark=et.get('mark')
	    #print_dip=True
	if ef.get('mark')!='':
	    mark=ef.get('mark')
	    #print_sip=True
	    
	if icmptype!='':
	    t.extend(['--icmp-type',icmptype])
	if dport!='':
	    t.extend(['--dport',dport])
	if sport!='':
	    t.extend(['--sport',sport])
	    
	#do we need to print the src/dst ip or dst/src net?
	if ef.get('numnet')!='':
	    sip=ef.get('numnet')
	if et.get('numnet')!='':
	    dip=et.get('numnet')
	if sip!='' and sip.find('/')==-1:
	    #t.extend(['-s',sip])
	    print_sip=True
	if dip!='' and dip.find('/')==-1:
	    #t.extend(['-d',dip])
	    print_dip=True
	if sip.find('/')!=-1:
	    if self._net_is_inclued_in(sip,self.netsrc) is True:
		#print 'adding sip',sip,self.rule.aclid
		#t.extend(['-s',sip])
		print_sip=True
	if dip.find('/')!=-1:
	    if self._net_is_inclued_in(dip,self.netdst) is True:
		#print 'adding dip',dip,self.rule.aclid
		#t.extend(['-d',dip])
		print_dip=True

	if mark!='' and self.do_vpn is True:
	    t.extend(['-m mark','--mark',mark])
	if print_sip is True:
	    t.extend(['-s',sip])
	if print_dip is True:
	    t.extend(['-d',dip])

	t.extend([self.add,'-j',self.target])
	if self.target==D_ULOG and self.ulog_prefix!='':
	    t.extend(['--ulog-prefix','"%s"'%(self.ulog_prefix)])
	if self.target==D_LOG and self.ulog_prefix!='':
	    t.extend(['--log-prefix','"%s "'%(self.ulog_prefix)])
	t.extend([' #',str(self.rule.aclid)])
	t.extend([str(self.rule.cpt)])
	
	s2=gen_rule2(t)
	return s2

