#! /usr/bin/env python

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


__doc__ = """Generates iptables commands and ldap nufw acls from xml file.

Nupyf takes an nufw acls xml file and an nufw firewall xml file and then:
    * output iptables commands
    * inserts acls into ldap tree

usage: nupyf [options] acls.xml firewall.xml

"""


__version__ = '0.9'
__author__ = 'Jean Gillaux'
__copyright = 'Copyright 2005, INL'

import xml.dom.minidom
import sys
from nuacl import *
import fw
import ipt
from os import linesep, times
from optparse import OptionParser
from time import strftime
from getpass import getpass
import nuldap
from ldap import LDAPError
import cPickle

def check_ldap(options):
    c = 0
    error = 0
    if options.ldap_server !='': c+=1
    if options.ldap_basedn != '': c+=1
    if c!=0 :
	if c!=2: 
	    print >> sys.stderr, 'Error: --server and --basedn options must be provided together'
	    sys.exit(1)
	if options.ldap_basedn =='':
	    print >> sys.stderr, 'Error: basedn must not be empty.'
	    sys.exit(1)

def parse_config_file(config_file,options):
    """Parse config_file and put directives into options.
    """
    try:
	config = eval(open(config_file).read())
    except IOError, ioe:
	print ioe
	sys.exit(1)
    for k,v in config.iteritems():
	if k == 'ldap_password': options.ldap_pwd = v
	if k == 'ldap_user': options.ldap_user = v
	if k == 'ldap_server': options.ldap_server = v
	if k == 'ldap_basedn': options.ldap_basedn = v
	
def main(argv):
    usage = "usage: %prog [options] firewall_file.xml acls_file.xml"
    parser = OptionParser(usage,version='%prog '+ __version__)
    parser.add_option('','--nufw',dest='nufw',help='file to output nufw iptables commands, special value - means stdout.',
			metavar='FILE',default='')
    parser.add_option('','--std',dest='std',help='file to output standard iptables commands (backup mode), special value - means stdout.',
			metavar='FILE',default='')
    parser.add_option('','--vpn',dest='vpn_rules',help='file to output vpn rules for packet marks, special value - means stdout',
			metavar='FILE',default='')
    parser.add_option('','--ulog',dest='ulog',action="store_true",help='use ULOG system in log commands, default is to use LOG.')
    parser.add_option('','--iptables',dest='ipt',help='path to iptables, can be a variable.',metavar='VALUE',default='')
    parser.add_option('-s','--server',dest='ldap_server',help='ldap server address.',metavar='ADDR',default='')
    parser.add_option('-u','--user',dest='ldap_user',help='ldap user used to connect.',metavar='USER',default='')
    parser.add_option('-p','--pwd',dest='ldap_pwd',help='password used to connect on ldap server (WARNING: visible on comand line).',
			metavar='PWD',default='')
    parser.add_option('-b','--basedn',dest='ldap_basedn',help='basedn used to insert acls in (ex: ou=acls,dc=domain,dc=com).',
			metavar='DN',default='')
    parser.add_option('-a','--askpwd',dest='ask_ldappwd',help='ask user for ldap passwd.',action='store_true',default=False)
    parser.add_option('','--config',dest='config_file',help='configuration file',default='')
    parser.add_option('','--dumpldap',dest='dump_ldap',help='file to save ldap informations',default='')
    parser.add_option('','--loadldap',dest='load_ldap',help='file to load ldap informations from',default='')
    
    parser.set_defaults(nufw='',std='',vpn_rules='', ulog=False, ipt='', ldap_server='', ldap_user='',
			lpdap_pwd='', ldap_basedn='', ask_ldappwd=False, config_file='', dump_ldap='', load_ldap='')
    (options,args)=parser.parse_args(argv)
    if options.config_file != '':
	parse_config_file(options.config_file,options)
    #print 'ldap: ',options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn
    #sys.exit(0)
    if options.ask_ldappwd is True:
	options.ldap_pwd = getpass('LDAP Password: ')

    do_ldap = False
    check_ldap(options)
    if options.ldap_server !='' and options.ldap_basedn != '':
	do_ldap = True
    if options.load_ldap != '':
	dapfw = cPickle.load(open(options.load_ldap))
	ldapconn=nuldap.LDAPConn(options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn)
	dapfw.set_conn(ldapconn)
	dapfw.load_ldap()
	dapfw.process()
	return 0
    
    if options.ipt!='': 
	#print 'setting ipt'
	ipt.IPTABLES=options.ipt
	#print ipt.IPTABLES
    do_stdout=0
    #no outputfile: we print the nufw iptables commands on stdout
    if options.std=='' and options.nufw=='': do_stdout=1
    log_type='log'
    if options.ulog is True: log_type='ulog'
        
    doc = xml.dom.minidom.parse(args[2])
    lr = doc.getElementsByTagName("ressource")
    ls = doc.getElementsByTagName("subject")
    lp = doc.getElementsByTagName("protocol")
    listr=[]
    listrobj=EltGrpList()
    lists=[]
    listsobj=EltGrpList()
    listp=[]
    listpobj=EltGrpList()

    for r in lr:
        #print r
	elts=r.getElementsByTagName("elt")
        #print elts
	try:
	    mygrp=xml_new_eltgrp(r.getAttribute('ID'),r.getAttribute('name'),'or',elts)
	except ValueError, e:
	    print e
	    return 1
        #print 'new ressource ',mygrp
	listr.append(mygrp)
        listrobj.add(mygrp)

    for s in ls:
	elts2=s.getElementsByTagName("elt")
	try:
	    mygrp2=xml_new_eltgrp(s.getAttribute('ID'),s.getAttribute('name'),s.getAttribute('op'),elts2)
	except ValueError, e:
	    print e
	    return 1
	lists.append(mygrp2)
        listsobj.add(mygrp2)

    for pr in lp:
	elts3=pr.getElementsByTagName("elt")
	try:
	    mygrp3=xml_new_eltgrp(pr.getAttribute('ID'),pr.getAttribute('name'),'or',elts3)
	except ValueError, e:
	    print e
	    return 1
	listp.append(mygrp3)
        listpobj.add(mygrp3)
    
    #listsobj.make_graph()
    #listpobj.make_graph()
    #listrobj.make_graph()
    #do the operator and link things
    #Shazam
    listsobj.normall()
    listrobj.normall()
    listpobj.normall()
    
    
    lacls=doc.getElementsByTagName("acl")
    l_acls=[]
    myfw=fw.parse(args[1])['1']
    for a in lacls:
	src=a.getAttribute("from")
        dst=a.getAttribute("to")
	proto=a.getAttribute("proto")
	aclname=a.getAttribute("name")
	#print src+"->"+dst+" with "+proto
        acl=Acl(listsobj.find(src),listrobj.find(dst),myfw,id=a.getAttribute('ID'),proto=listpobj.find(proto),name=aclname,decision=a.getAttribute('decision'))
	#for r in acl.rules():
	    #myfw.add_rule(r)	

	l_acls.append(acl)
    
    #net_ent=myfw.entreprise_nets()
    myfw.chains(all=1)
    
    for x in l_acls: 
	for r in x.rules():
	    myfw.add_rule(r)	
    
    #create iptables commands
    fwp=ipt.FWipt(myfw,logtype=log_type)
    rules_create = fwp.create_ipt_chains()
    rules_connect = fwp.connect_chains()

    
    intro="""
#Generated by nupyf on %s

""" %(strftime('%d %m %Y at %H:%M'))
    if do_stdout==1 or options.nufw != '' :
	if options.nufw=='-':
	    do_stdout=1
	if do_stdout!=1:
	    fd=open(options.nufw,'w')
	else: fd=sys.stdout
	srules = fwp.gen_rules()
	fd.write(intro)
	fd.write(rules_create)
	fd.write(rules_connect)
	#fd.write(srules)
	fd.writelines(srules)
	if do_stdout!=1:
	    fd.close()
    if options.std != '':
	do_stdout=0
	if options.std=='-':
	    do_stdout=1
    	rules_rescue= fwp.gen_rules(rescue=1)
	if do_stdout!=1:
	    fd=open(options.std,'w')
	else: fd=sys.stdout
	fd.write(intro)
	fd.write(rules_create+linesep)
	fd.write(fwp.connect_chains(rescue=1)+linesep)
	#fd.write(rules_rescue)
	fd.writelines(rules_rescue)
	if do_stdout!=1:
	    fd.close()
    #vpn mark rules
    if options.vpn_rules != '':
	do_stdout = 0
	if options.vpn_rules != '-':
	    do_stdout = 1
	    fd = open(options.vpn_rules,'w')
	else: fd = sys.stdout
	fd.writelines(fwp.gen_vpn_rules())
	if do_stdout != 1: fd.close()
    #LDAP stuff
    if do_ldap is True:
	conn = None
	ldapfw=nuldap.LDAPfw(conn,fw=myfw)
	if options.dump_ldap != '':
	    cPickle.dump(ldapfw,open(options.dump_ldap,'w'))
	    return 0
	conn=nuldap.LDAPConn(options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn)
	ldapfw.set_conn(conn)
	ldapfw.load_ldap()
	ldapfw.process()
    return 0
    

if __name__ == "__main__":
    try:
    	res=main(sys.argv)
    except LinkError, linke:
	print >> sys.stderr, str(linke)
	sys.exit(1)
    except LoopError, loope:
	print >> sys.stderr, str(loope)
	sys.exit(1)
    except fw.UnknownVpn, evpn:
	print >> sys.stderr, 'ERROR: unknown vpn with mark %s detected!'%(evpn.mark)
	sys.exit(1)
    except LDAPError, ldaperr:
	print >> sys.stderr, 'ERROR ldap:', ldaperr
	sys.exit(1)
    except ValueError, ve:
	print 'main intercept ValueError'
	print str(ve)
	sys.exit(1)
    sys.exit(res)
